"""
    Genetically modified black cat
"""
import numpy as np
import random as r
import math


"""
===============================================================================
    Genome Class !@#$%^&*
===============================================================================
"""
class aCat():
    def __init__(self, how_many_genes, r_min, r_max, mom=0.0, dad=0.0):
        self.name = r.random()
        self.fitness = 0
        self.active = True
        self.how_many_genes = how_many_genes
        self.genes = self.randomGenes(how_many_genes, r_min, r_max)
        # Family
        self.parents = [mom, dad]

    """
        Create random genomes
    """
    def randomGenes(self, n, r_min, r_max):
        return [r.uniform(r_min, r_max) for _ in range(n)]

"""
===============================================================================
    Our Black Cats class !@#$%^&*
===============================================================================
"""
class newCats():
    def __init__(
                    self, # Self
                    action_space, # Action space
                    population_size, # How many genoms
                    genes_in_genome, # How many genes in genom
                    elite_alives, # How many best genoms should stay alive
                    mutation_rate, # How genoms will mutate
                    mutation_step, # How much genoms will mutate
                    random_cats=0, # How many random cats should be in population
                    random_min_gene=(-0.5),
                    random_max_gene=0.5,
                    min_gene_val=None,
                    max_gene_val=None
                ):

        self.id = r.random()
        self.action_space = action_space
        self.population_size = population_size
        self.genes_in_genome = genes_in_genome
        self.mutation_rate = mutation_rate
        self.mutation_step = mutation_step
        self.all_cats = []
        self.current_cat = 0
        self.elite_alives = elite_alives
        self.generation = 0
        self.random_cats = random_cats
        self.random_min_gene = random_min_gene
        self.random_max_gene = random_max_gene
        self.min_gene_val = min_gene_val
        self.max_gene_val = max_gene_val
        self.actual_state = None

        # Time to init our genomes
        self.bornCats()

    """
        Set current genome
    """
    def setCurrentCat(self, gen):
        self.current_cat = gen

    """
        Return object ID
    """
    def familyName(self):
        return self.id

    """
        Return current genome number
    """
    def currentCat(self):
        return self.current_cat

    """
        Return cat object
    """
    def cuddleCat(self):
        return self.all_cats[self.current_cat]

    """
        Return ID's of cat parents
    """
    def currentCatParents(self):
        return self.all_cats[self.currentCat].parents

    """
        Return current genome details:
            - ID
            - Genes
    """
    def currentCatDetail(self):
        return [
                    self.all_cats[self.current_cat].name,
                    self.all_cats[self.current_cat].genes
                ]

    def addPoint(self, point = 1):
        self.all_cats[self.current_cat].fitness += point

    def genomeGameOver(self):
        self.all_cats[self.current_cat].active = False

    def sigmoid(self, x):
        return 1 / (1 + math.exp(-x))

    """
        Initialize first genomes (randomly ofc)
    """
    def bornCats(self):
        for _ in range(self.population_size):
            self.all_cats.append(aCat(
                                        self.genes_in_genome,
                                        self.random_min_gene,
                                        self.random_max_gene
                                    ))

    def rateAction(self, observation, reward):
        # Init rate
        rate = 0
        # Loop genes
        for i in range(0, self.genes_in_genome):
            rate += self.all_cats[self.current_cat].genes[i] * observation[i]
        # Include reward
        rate += 500 if reward > 0 else -500
        # Return
        return rate

    def chooseAction(self, action_rates):
        return action_rates.index(max(action_rates))

    def action(self, env):
        action_rates = []
        for action in self.action_space:
            observation, reward, _, _ = env.step(action)
            action_rates.append(self.rateAction(observation, reward))

        preffered_action = self.chooseAction(action_rates)

        return preffered_action

    def resetFitness(self):
        for i in range(self.population_size):
            self.all_cats[i].fitness = 0

    def sortFitness(self):
        self.all_cats = sorted(self.all_cats, key=lambda k: k.fitness, reverse=True)

    def mutateGenome(self, parents):
        # Our parents
        mum, dad = parents
        # New genome
        new_genome = aCat(self.genes_in_genome, self.random_min_gene, self.random_max_gene)
        # Mutate this genome with mum and dad
        for i, gen in enumerate(new_genome.genes):
            new_genome.genes[i] = r.choice([mum.genes[i], dad.genes[i]])
            # Mutation
            if r.random() < self.mutation_rate:
                new_genome.genes[i] += (r.random() * self.mutation_step * 2) - self.mutation_step
                if self.min_gene_val != None and self.max_gene_val != None:
                    if new_genome.genes[i] < self.min_gene_val:
                        new_genome.genes[i] = self.min_gene_val
                    elif new_genome.genes[i] > self.max_gene_val:
                        new_genome.genes[i] = self.max_gene_val
        # Born
        return new_genome

    def nextPopulation(self):
        # Sort genomes by fitness
        self.sortFitness()

        # Create list of elite and new genomes
        _elite_alives = []
        _elite_weights = []
        _all_cats = []
        _random_cats = []

        # Append elite's
        for i in range(self.elite_alives):
            _elite_alives.append(self.all_cats[i])
            if self.all_cats[i].fitness > 0:
                _elite_weights.append(self.all_cats[i].fitness / self.all_cats[0].fitness)
            else:
                _elite_weights.append(0.001)

        # Mutate genomes and add to list
        for i in range(self.elite_alives, self.population_size - self.random_cats):
            _all_cats.append(self.mutateGenome(r.choices(
                population=_elite_alives,
                weights=_elite_weights,
                k=2
            )))

        # Create some new random cats
        for i in range(self.population_size - self.random_cats, self.population_size):
            _random_cats.append(aCat(self.genes_in_genome, self.random_min_gene, self.random_max_gene))

        self.all_cats = _elite_alives + _all_cats + _random_cats
        r.shuffle(self.all_cats)
        self.resetFitness()
